#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# SPDX-License-Identifier: Apache-2.0
# Copyright 2024 Inria
#
## \namespace upkie.model
## \brief Robot model.

from .model import Model

__all__ = [
    "Model",
]
